package demux.ty;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Map;

import muxer.MuxManager;
import muxer.PacketBuffer;
import muxer.PacketBufferRegulatorI;
import muxer.pes.PesMuxer;

import common.BufferedWritableByteChannel;
import common.FileStream;
import common.Job;
import common.JobStreamI;
import common.Logger;
import common.PropertiesReader;
import common.StreamIdentifier;
import common.Util;
import common.VideoStreamDescriptor;
import common.mpeg.MpegConstants;
import common.mpeg.MpegPacketSourceBuffer;
import common.mpeg.MpegUtil;
import common.mpeg.PesPacketInfo;
import common.mpeg.SequenceHelper;
import common.packet.Packet;
import common.packet.PacketPart;
import common.packet.PesPacket;
import common.packet.PesPacketPart;

import demux.Demuxer;

public class TyPesDemuxer extends Demuxer implements MpegConstants {
	private static final byte TIVO_PES_FILEID[]={(byte)0xf5,0x46,(byte)0x7a, (byte)0xbd};
	private static final int TY_CHUNK_SIZE=128*1024;

	private static final String BAD_TY_TIMESTAMP_THRESHOLD_PROPERTY="TyStreamSource.badTyTimestampThreshold";
	
	private Logger logger=Logger.getLogger(this.getClass());
	//private byte[] buffer= new byte[TY_CHUNK_SIZE];
	private long badTyTimestampThreshold;
	private MpegPacketSourceBuffer video=new MpegPacketSourceBuffer();
	private MpegPacketSourceBuffer audio=new MpegPacketSourceBuffer();
	
	private TyStatistics cachedStats;
	//private int zVideoBytes=0;
	
	private interface PacketSanitizerI {
		public void sanitize(MpegPacketSourceBuffer buffer, PesPacketInfo info);
	}
	
	private final PacketSanitizerI nullSanitizer= new PacketSanitizerI() {
		public void sanitize(MpegPacketSourceBuffer buffer, PesPacketInfo info) {}
	};
	private final PacketSanitizerI mpeg2Sanitizer = new PacketSanitizerI () {
		public void sanitize(MpegPacketSourceBuffer buffer, PesPacketInfo info) {
			int packetEnd=info.getEnd();
			byte myByte=video.get(packetEnd - 2);
			while (myByte == (byte)0x80 || myByte==(byte)0) {
				myByte=video.get((--packetEnd) - 2);
			}
			info.setEnd(packetEnd);
		}
	};
	
	class TyStreamIdentifierRegulator implements PacketBufferRegulatorI {
		private SequenceHelper sequenceHelper=null;
		private int audioStreamId;
		
		private int index=-1;
		
		public void addedPacket(PacketBuffer packetBuffer, Packet lastPacket) {
			if (audioStreamId==0) {
				Packet dummy=packetBuffer.getFirstPacket(0xc0);
				if (dummy==null) {
					dummy=packetBuffer.getFirstPacket(0xbd);
				}
				if (dummy != null) {
					audioStreamId=dummy.getStreamId();
				}
			}
			int lastIndex=packetBuffer.getLastIndex();
			if (index==-1) {
				index=lastPacket.getIndex();
			}
			while (index <= lastIndex) {
				Packet packet=packetBuffer.getPacket(index);
				if (packet == null) {
					//indicates a packet is buffering, we'll wait for it
					break;
				} else {
					index++;
					if (packet.getStreamId()==0xE0) {
						PacketPart data=packet.getPayload();
						int startCode=MpegUtil.findStartCode(MpegConstants.SEQUENCE_PACKET, data.getByteBuffer(), 0);
						if (startCode>=0) {
							sequenceHelper= new SequenceHelper(data, startCode);
						}
					}
				}
			}
		}
		public int isFull() {
			if (sequenceHelper != null && audioStreamId !=0) {
				return MuxManager.ABORT;
			}
			return MuxManager.DO_NOTHING;
		}
		final public void removedPacket(Packet packet){}
	}
	
	interface ByteSourceI {
/*		public void skip(long bytes) throws IOException;*/
		public ByteBuffer read() throws IOException;
	}
	

	public TyPesDemuxer() {
		badTyTimestampThreshold=PropertiesReader.getInstance().getLongProperty(BAD_TY_TIMESTAMP_THRESHOLD_PROPERTY, 3000);
	}
	
	
	public static void main(String args[]) {
		try {
			String fileName="Tombstone";
			System.out.println("Starting");
			//String fileName="Grey's Anatomy - Some Kind of Miracle";
			//String fileName="chunk";
			String titleName="H:\\test Tys\\"+fileName+".ty";
			//String titleName="C:\\Tivo\\chunk.mpg";
			//InputStream in=new FileInputStream(titleName);
			FileStream fileJob=new FileStream(new Job(null), new File[]{new File(titleName)});
			TyPesDemuxer demuxer=new TyPesDemuxer();
			//new RestreamingPesMuxer(buffer, new NullOutputStream());
			long start=System.currentTimeMillis();
			//in.skip(10631512064l);
			MuxManager muxManager=new MuxManager(demuxer);
			Map map=muxManager.getStreamDescriptorMap(fileJob);
			System.out.println(map);
			muxManager.add(new PesMuxer(new BufferedWritableByteChannel(new File("C:/test.mpg"),10*1024*1024)));
			muxManager.mux(fileJob);
			System.out.println("Done: " + (System.currentTimeMillis()-start));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Figures out if this is a part header
	 * 
	 * @param buffer
	 * @param offset
	 * @return
	 */
	public static boolean isTivoPartId(ByteBuffer chunk) {
		for (int i=0; i<TIVO_PES_FILEID.length;i++) {
			if (chunk.get(i) != TIVO_PES_FILEID[i]) {
				return false;
			}
		}
		return true;
	}
	
	
	public void parse(JobStreamI stream, MuxManager muxManager) throws IOException {
		stream.setChunkSize(TY_CHUNK_SIZE);
		//Since this is designed to be called re-entranly only open the file once
		if (muxManager.getPacketBuffer().size()==0) {
			stream.open();
		}
		cachedStats=parse(stream, muxManager, cachedStats==null ? new TyStatistics() : cachedStats);
	}
	public TyStatistics parse(JobStreamI stream, MuxManager muxManager, TyStatistics stats) throws IOException {
		
		//byte chunk[]=new byte[TY_CHUNK_SIZE];
		ByteBuffer chunk;
		ChunkHeader chunkHeader=new ChunkHeader();
		RecordHeader recordHeader=new RecordHeader();
		//int counter=0;
		//Use processing flag so we always process a complete chunk
		boolean processing=true;
		//ByteBuffer chunk;
		while (!stream.isEOS() && processing) {
			//stream.readFully(chunk, 0, TY_CHUNK_SIZE);
			chunk=stream.slice(TY_CHUNK_SIZE);
			/*if (chunk==null) {
				break;
			}*/
			//System.out.println("Chunk: " + (counter++));
			boolean isPartHeader=TyPesDemuxer.isTivoPartId(chunk);
			
			if (stats.getChunkCounter()==0 && !isPartHeader) {
				throw new IOException("Encountered invalid Ty Stream.  Did you disable encryption?");
			}
			stats.incrementPackets();
			//System.out.println(stats.getChunkCounter());
			//counter++;
			if (isPartHeader) {
				continue;
			}
			
			int recordOffset=chunkHeader.parse(chunk);
			int dataOffset=chunkHeader.getDataOffset();
			
			for (int i=0; i<chunkHeader.records; i++) {
				recordHeader.parse(chunk, recordOffset);
				
				if (stats.getLastTimeStamp()!=0) {
					long delta=Math.abs(recordHeader.timestamp-stats.getLastTimeStamp())/Util.TIVO_TIMESCALE_MILLIS;
					if (delta > badTyTimestampThreshold) {
						//This is very normal
						if (logger.isTrace()) {
							logger.trace("Aborting record due to large timestamp delta: " + delta + " | " + Util.tivoTimeToString(recordHeader.timestamp) + " | " + Util.tivoTimeToString(stats.getLastTimeStamp()));
						}
						break;
					}
				}
				int recordSize=getRecordSize(recordHeader, dataOffset);
				if (recordSize > 0) {
					if (recordSize + dataOffset <= TY_CHUNK_SIZE) {
						//System.out.println("Timestamp " + Util.tivoTimeToString(recordHeader.timestamp));
						
						processing &=(processRecord(recordHeader, recordSize, chunk, dataOffset, muxManager)==MuxManager.CONTINUE);
						dataOffset +=recordSize;
					} else { 
						logger.warn("Invalid Record Size Returned: " + recordSize + " : " + recordHeader.type);
					} 
				}
				stats.setLastTimeStamp(recordHeader.timestamp);
				recordOffset+=16;
			}

		}
		if (processing) {
			muxManager.streamEnded();
		}
		return stats;
	}
	
	private int getRecordSize(RecordHeader recordHeader, int dataOffset) {
		byte type=recordHeader.type;
		int size=recordHeader.recordSize;
		if (type==0x01 || type==0x02 || type==0x0E) {
			//System.out.println("type: " + type + ": " +typeNibble+": " + size);
			return 0;
		}
		
		if (size + dataOffset <= TY_CHUNK_SIZE) {
			return size;
		} else {
			//System.out.println("Illegal size for packet type: " + type + ": " + size);
			logger.debug("Illegal size for packet type: " + type + ": " + size);
			return 0;
		}
	}
	
	private int processRecord(RecordHeader recordHeader, int recordSize, ByteBuffer chunk, int dataOffset, MuxManager muxManager) throws IOException {
		byte type=recordHeader.type;
		byte typeNibble=recordHeader.typeNibble;
		int rtn;
		if (type==VIDEO_PACKET) {
			video.appendFrom(chunk, dataOffset, recordSize);
			rtn=demux(video, mpeg2Sanitizer, muxManager);
		} else if (type==AUDIO_PACKET) {
			//AC3==9
			if (typeNibble==2 || typeNibble==3 || typeNibble==4 || typeNibble==9) {
				audio.appendFrom(chunk, dataOffset, recordSize);
				rtn=demux(audio, nullSanitizer, muxManager);
			} else {
				System.out.print("Uknown Audio Type: ");
				Util.printHex(type);
				System.out.println();
				rtn=MuxManager.CONTINUE;
			}
		} else {
			//Do nothing with unknown types
			rtn=MuxManager.CONTINUE;
		}
		return rtn;
	}
	private int demux(MpegPacketSourceBuffer buffer, PacketSanitizerI packetSanitizer, MuxManager muxManager) throws IOException {
		if (buffer.limit() < 16) {
			return MuxManager.CONTINUE;
		}
		int rc=MuxManager.CONTINUE;
		for (PesPacketInfo info;(info=buffer.getPesPacket()).isComplete() && rc==MuxManager.CONTINUE;) {
			final int end=info.getEnd();
			packetSanitizer.sanitize(buffer, info);
			
			final int start=info.getStart();
			PesPacket packet = new PesPacket(buffer.get(start + 3) & 0xff,info.getIndex());

			PesPacketPart header;
			if (MpegUtil.getMpegType(buffer, start)==2) {
				int len=info.getData()-start;
				header=muxManager.getPesPacketPart(len);
				header.appendFrom(buffer.getByteBuffer(), start, len);
			} else {
				header=MpegUtil.upgradeMpeg1PesHeader(buffer, start);
				//Hack
				info.setData(start + header.limit() - 3);
			}
			packet.setHeader(header);
			int len=info.getEnd()-info.getData();
			PacketPart payload=muxManager.getPacketPart(len);
			payload.appendFrom(buffer.getByteBuffer(), info.getData(), len);
			packet.setPayload(payload);
			rc=muxManager.addPacket(packet);
			buffer.discard(end);
		}
		return rc;
	}
	public Map getStreamDescriptorMap(JobStreamI stream, MuxManager muxManager) throws IOException {
		TyStreamIdentifierRegulator regulator=new TyStreamIdentifierRegulator();

		PacketBuffer packetBuffer=new PacketBuffer(regulator); 
		muxManager.add(packetBuffer);
		muxManager.mux(stream);
		HashMap map=new HashMap();
		VideoStreamDescriptor descriptor=new VideoStreamDescriptor(0xE0);
		descriptor.setCodec("MPEG-2");
		descriptor.setFrameRateRatio(MpegUtil.getFrameRateRatio(regulator.sequenceHelper.getFrameRateType()));
		descriptor.setAspectRatio(MpegConstants.ASPECT_RATIO[regulator.sequenceHelper.getAspectRatioType()]);
		descriptor.setHeight(regulator.sequenceHelper.getHeight());
		descriptor.setWidth(regulator.sequenceHelper.getWidth());
		descriptor.setTsType(TS_STREAM_TYPE_MPEG2_VIDEO);
		map.put(Integer.valueOf(descriptor.getStreamId()),descriptor);
		
		if (regulator.audioStreamId==0xc0) {
			map.put(Integer.valueOf(regulator.audioStreamId),new StreamIdentifier.MpegAudioAnalyzer().analyze(packetBuffer.getFirstPacket(regulator.audioStreamId).getPayload(), regulator.audioStreamId, 0));
		} else {
			map.put(Integer.valueOf(regulator.audioStreamId),new StreamIdentifier.DolbyDigitalAnalzyer().analyze(packetBuffer.getFirstPacket(regulator.audioStreamId).getPayload(), regulator.audioStreamId, 0));
		}
		packetBuffer.removeRegulator(regulator);
		
		return map;
	}
}
