package util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;

import muxer.MuxFilter;
import muxer.MuxManager;
import muxer.MuxManager.DemuxerMuxerInfo;
import muxer.demux.RawDemuxerMuxer;
import ui.CliJobListener;

import common.AudioStreamDescriptor;
import common.Job;
import common.Logger;
import common.StreamDescriptor;
import common.VersionI;
import common.VideoStreamDescriptor;

import demux.Demuxer;
import demux.microsoft.WtvFileStream;
import demux.microsoft.WtvToPesDemuxer;

public class WtvToMpeg implements VersionI {
	private static Logger logger=Logger.getLogger(WtvToMpeg.class);
	
	public static final String HELP= "WtvToMpeg "+FULL_VERSION + "\n" +
								"Usage:\n" +
								" [-i] inputFile.wtv\n" +
								" [-o - | fileName.[mpg | ts | m2ts]\n" +
								" [-lang eng|fra|ger|...]\n" +
								" [-all]\n" +
								" [-demux]\n\n" +
							    " 'fileName.wtv' will create 'fileName.ts'\n" +
							    " '-o -' will output to stdout\n" +
							    " '-demux' will demux to fileName.mpv, fileName.ac3, etc.\n" +
							    " '-all' will keep all audio tracks";
	
	public static void main(String args[]) {
		if (args.length==0) {
			System.out.println(HELP);
			System.exit(-1);
		}
		
		String inFileName=null;
		String outFileName=null;
		boolean demux=false;
		String lang=null;
		boolean allAudio=false;
		
		for (int i=0; i < args.length; i++) {
			if (args[i].equals("-i")) {
				inFileName=args[++i];
			} else if (args[i].equals("-o")) {
				outFileName=args[++i];
			} else if (args[i].equals("-lang")) {
				lang=args[++i];
			} else if (args[i].equals("-all")) {
				allAudio=true;
			} else if (args[i].equals("-demux")) {
				demux=true;
			} else if (i==0) {
				inFileName=args[i];
			}
		}
		if (demux) {
			if (inFileName==null) {
				System.out.println("Input File required");
				System.out.println(HELP);
				System.exit(-1);
			} else if (outFileName != null && "-".equals(outFileName)) {
				System.out.println("Can not demux to stdOut");
				System.out.println(HELP);
				System.exit(-1);
			}
		}
		File in=new File(inFileName);
		if (!in.getAbsolutePath().toLowerCase().endsWith(".wtv")) {
			System.out.println("This utility only works with WTV files");
			System.out.println(HELP);
			System.exit(-2);
		}
		String inFileSuffix=inFileName.substring(0, inFileName.length() -4);
		if (outFileName==null) {
			if (demux) {
				outFileName= inFileSuffix;
			} else {
				outFileName= inFileSuffix + ".ts";
			}
		}
		Demuxer demuxer=new WtvToPesDemuxer();
		MuxManager muxManager=new MuxManager(demuxer);
		
		boolean pipe=outFileName.equals("-");
		CliJobListener cliListener=pipe ? null : new CliJobListener();
		Job job=new Job(cliListener);
		
		WtvFileStream stream=new WtvFileStream(new Job(), in);
		try {
			Map<Integer, StreamDescriptor> streamMap=muxManager.getStreamDescriptorMap(stream);
			
			VideoStreamDescriptor vsd=null;
			AudioStreamDescriptor asd=null;
			ArrayList<AudioStreamDescriptor> audioList=new ArrayList<AudioStreamDescriptor>();
			for (StreamDescriptor sd : streamMap.values()) {
				if (sd instanceof VideoStreamDescriptor) {
					vsd=(VideoStreamDescriptor)sd;
				} else if (sd instanceof AudioStreamDescriptor) {
					AudioStreamDescriptor asdTemp=(AudioStreamDescriptor)sd;
					audioList.add(asdTemp);
					if (asd==null) {
						asd=asdTemp;
						continue;
					} else if (lang != null) {
						if (lang.equals(asd.getLanguage())) {
							if (!lang.equals(asdTemp.getLanguage())) {
								//Current ASD is the preferred language
								//and temp is not
								continue;
							}
						} else if (lang.equals(asdTemp.getLanguage())) {
							//Current ASD is not preferred language 
							//and the temp is
							asd=asdTemp;
							continue;
						}
					} 
					if (asdTemp.getBitRate() > asd.getBitRate()) {
						asd=asdTemp;
					}
				}
			}
			ArrayList<StreamDescriptor> sdList=new ArrayList<StreamDescriptor>(2);
			sdList.add(vsd);
			sdList.add(asd);
			
			if (allAudio) {
				audioList.remove(asd);//Don't double up
				sdList.addAll(audioList);
			}
			demuxer.setStreams(sdList);
			
			if (!pipe) {
				for (StreamDescriptor sd : streamMap.values()) {
					StringBuffer sb=new StringBuffer();
					if (sdList.contains(sd)) {
						sb.append('+');
					} else {
						sb.append('-');
					}
					sb.append(sd);
					logger.info(sb);
				}
			}
			MuxFilter muxer=null;

			if (demux) {
				muxer=new RawDemuxerMuxer(outFileName);
			} else {
				DemuxerMuxerInfo[] remuxInfo=MuxManager.getDemuxerRemuxerInfo(in);
				String extension=pipe ? ".ts" : outFileName.substring(outFileName.lastIndexOf('.'));
				for (DemuxerMuxerInfo info : remuxInfo) {
					if (info.getExt().equals(extension)) {
						muxer=MuxManager.createRemuxer(info.getRemuxerClass(), new File(outFileName));
						break;
					}
				}
				if (muxer==null) {
					System.out.println(HELP);
					System.out.println("Unknown output type " + extension);
				}
			}
			muxManager.add(muxer);
			stream=new WtvFileStream(job, in);
			muxManager.mux(stream);
			if (!pipe) {
				//Clear status line
				System.out.println();
				logger.info("Finished");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
